"""Commandline API gateway for laskea."""
import sys

import requests_cache
import typer

import laskea
import laskea.config as cfg
import laskea.laskea as fill
from laskea import env

app = typer.Typer(
    add_completion=False,
    context_settings={'help_option_names': ['-h', '--help']},
    no_args_is_help=True,
)

Version = typer.Option(
    False,
    '-V',
    '--version',
    help='Display the laskea version and exit',
    is_eager=True,
)

Input = typer.Option(
    '',
    '-i',
    '--input',
    help='Path to input file',
    metavar='<sourcepath>',
)

ConfigurationPath = typer.Option(
    '',
    '-c',
    '--config',
    help=f'Path to config file (default is $HOME/{laskea.DEFAULT_CONFIG_NAME})',
    metavar='<configpath>',
)

Dryness = typer.Option(
    False,
    '-n',
    '--dry-run',
    help='Dry run (default is False)',
)

Verbosity = typer.Option(
    False,
    '-v',
    '--verbose',
    help='Verbose output (default is False)',
)

Quietness = typer.Option(
    False,
    '-q',
    '--quiet',
    help='Minimal output (default is False)',
)

Strictness = typer.Option(
    False,
    '-s',
    '--strict',
    help='Ouput noisy warnings on console and in the processed document (default is False)',
)

CacheExpiry = typer.Option(
    180,
    '-x',
    '--cache-expiry-seconds',
    help='Request cache expiry in seconds',
)

JqlQuery = typer.Option(
    '',
    '-j',
    '--jql-query',
    help=(
        'The query in JQL format.'
        '\nFor example given a project YES and two issues 123 and 124:'
        "\n'project = YES and key in (YES-123, YES-124) order by created DESC'"
    ),
    metavar='<jql-query>',
)

KeyMagic = typer.Option(
    False,
    '-k',
    '--key-magic',
    help='Apply magic to key by replacing with markdown like link (default is False)',
)

FieldSeparator = typer.Option(
    laskea.PIPE,
    '-d',
    '--delimiter',
    help=(
        'Delimiter / field separator'
        '\nOn output, header and data cell values will have any occurences'
        '\nof the field separator replaced with the replacement string'
    ),
    metavar='<field-separator>',
)

Replacement = typer.Option(
    laskea.FS_SLUG,
    '-r',
    '--replacement',
    help=(
        'Replacement string for occurences of FS in text\n'
        '\nOn output, header and data cell values will have any occurences'
        '\nof the field separator replaced with the replacement string'
    ),
    metavar='<replacement-text>',
)


@app.callback(invoke_without_command=True)
def callback(version: bool = Version) -> None:
    """Calculate (Finnish: laskea) some parts."""
    if version:
        typer.echo(f'{laskea.APP_NAME} version {laskea.__version__}')
        raise typer.Exit()


@app.command('template')
def app_template() -> int:
    """
    Write a template of a well-formed JSON configuration to standard out and exit
    """
    sys.stdout.write(cfg.generate_template())
    return sys.exit(0)


@app.command('report')
def report(
    shallow: bool = typer.Option(
        False,
        '-s',
        '--shallow',
        help='Shallow reporting - no setuptools required (default is False)',
    ),
) -> int:
    """Output either text options for the user to report her env or the report of the environment for support."""
    sys.stdout.write(env.report(shallow))
    return sys.exit(0)


@app.command('update')
def update(  # noqa
    source: list[str],
    inp: str = Input,
    conf: str = ConfigurationPath,
    verify: bool = Dryness,
    verbose: bool = Verbosity,
    quiet: bool = Quietness,
    strict: bool = Strictness,
    expires: int = CacheExpiry,
) -> int:
    """
    Fill in some parts of the input document.

    You can set some options per evironment variables:

    \b
    * LASKEA_USER='remote-user'
    * LASKEA_TOKEN='remote-secret'
    * LASKEA_BASE_URL='https://remote-jira-instance.example.com/'
    * LASKEA_CACHE_EXPIRY_SECONDS=180
    * LASKEA_COL_FIELDS: '[
        "Key",
        "Summary",
        "Custom Field Name",
        ["Custom Field Other", "Display Name"]
      ]'
    * LASKEA_COL_MAPS='{
        "key": [
          "key",
          "key"
        ],
        "summary": [
          "summary",
          "fields.summary"
        ],
        "custom field name": [
          "customfield_11501",
          "fields.customfield_11501"
        ],
        "custom field other": [
          "customfield_13901",
          "fields.customfield_13901[].value"
        ]
      }'
    * LASKEA_COL_FILTERS='{
        "key": {},
        "summary": {},
        "custom field name": {
          "order": ["keep", "drop", "replace"],
          "keep": [
            ["startswith", "ABC-"],
            ["contains", "Z"],
            ["icontains", "m"],
            ["equals", "DEF-42"],
            ["endswith", "-123"]
          ],
          "drop": [
            ["matches", "[A-Z]+-\\d+"]
          ],
          "replace": [
            ["DEF-", "definition-"]
          ]
        },
        "custom field other": {}
      }'
    * LASKEA_JOIN_STRING=' <br>'
    * LASKEA_LF_ONLY='AnythingTruthy'
    * LASKEA_CAPTION='empty or special DSL'
    * LASKEA_IS_CLOUD='WhenNotConnectingToJiraServerButJiraCloud'
    * LASKEA_MARKERS='[[[fill ]]] [[[end]]]'
    * LASKEA_DEBUG='AnythingTruthy'
    * LASKEA_VERBOSE='AnythingTruthy'
    * LASKEA_STRICT='AnythingTruthy'

    The quiet option (if given) disables any conflicting verbosity setting.
    """
    command = 'update'
    transaction_mode = 'commit' if not verify else 'dry-run'
    if quiet:
        laskea.QUIET = True
        laskea.DEBUG = False
        laskea.VERBOSE = False
    elif verbose:
        laskea.VERBOSE = True

    if strict:
        laskea.STRICT = True

    if transaction_mode == 'dry-run':
        laskea.DRY_RUN = True

    requests_cache.install_cache(cache_name='.laskea_cache', backend='sqlite', expire_after=expires)
    laskea.CACHE_EXPIRY_SECONDS = expires
    options = {
        'quiet': quiet,
        'strict': strict,
        'verbose': verbose,
    }
    cfg.process(conf, options)

    paths = (inp,) if inp else tuple(source)
    return sys.exit(fill.process(command, transaction_mode, paths, options))


@app.command('csv')
def svl_cmd(  # noqa
    query: str = typer.Argument(''),
    jql_query: str = JqlQuery,
    conf: str = ConfigurationPath,
    key_magic: bool = KeyMagic,
    field_sep: str = FieldSeparator,
    replacement: str = Replacement,
    verify: bool = Dryness,
    verbose: bool = Verbosity,
    strict: bool = Strictness,
    expires: int = CacheExpiry,
) -> int:
    """
    Export query result as separated values list.

    You can set some options per evironment variables:

    \b
    * LASKEA_USER='remote-user'
    * LASKEA_TOKEN='remote-secret'
    * LASKEA_BASE_URL='https://remote-jira-instance.example.com/'
    * LASKEA_CACHE_EXPIRY_SECONDS=180
    * LASKEA_COL_FIELDS: '[
        "Key",
        "Summary",
        "Custom Field Name",
        ["Custom Field Other", "Display Name"]
      ]'
    * LASKEA_COL_MAPS='{
        "key": [
          "key",
          "key"
        ],
        "summary": [
          "summary",
          "fields.summary"
        ],
        "custom field name": [
          "customfield_11501",
          "fields.customfield_11501"
        ],
        "custom field other": [
          "customfield_13901",
          "fields.customfield_13901[].value"
        ]
      }'
    * LASKEA_COL_FILTERS='{
        "key": {},
        "summary": {},
        "custom field name": {
          "order": ["keep", "drop", "replace"],
          "keep": [
            ["startswith", "ABC-"],
            ["contains", "Z"],
            ["icontains", "m"],
            ["equals", "DEF-42"],
            ["endswith", "-123"]
          ],
          "drop": [
            ["matches", "[A-Z]+-\\d+"]
          ],
          "replace": [
            ["DEF-", "definition-"]
          ]
        },
        "custom field other": {}
      }'
    * LASKEA_JOIN_STRING=' <br>'
    * LASKEA_LF_ONLY='AnythingTruthy'
    * LASKEA_CAPTION='empty or special DSL'
    * LASKEA_IS_CLOUD='WhenNotConnectingToJiraServerButJiraCloud'
    * LASKEA_MARKERS='[[[fill ]]] [[[end]]]'
    * LASKEA_DEBUG='AnythingTruthy'
    * LASKEA_VERBOSE='AnythingTruthy'
    * LASKEA_STRICT='AnythingTruthy'

    The quiet option (if given) disables any conflicting verbosity setting.
    """
    transaction_mode = 'commit' if not verify else 'dry-run'
    jql = jql_query.strip()
    if not jql and query:
        jql = query
    if not jql:
        print('JQL query required.', file=sys.stderr)
        return sys.exit(2)
    quiet = True
    laskea.QUIET = True
    laskea.DEBUG = False
    laskea.VERBOSE = False
    if verbose:
        laskea.VERBOSE = True

    if strict:
        laskea.STRICT = True

    if transaction_mode == 'dry-run':
        laskea.DRY_RUN = True

    requests_cache.install_cache(cache_name='.laskea_cache', backend='sqlite', expire_after=expires)
    laskea.CACHE_EXPIRY_SECONDS = expires
    options = {
        'quiet': quiet,
        'strict': strict,
        'verbose': verbose,
    }
    if conf:
        cfg.process(conf, options)

    return sys.exit(laskea.svl(jql, key_magic=key_magic, field_sep=field_sep, replacement=replacement))


@app.command('version')
def app_version() -> None:
    """
    Display the laskea version and exit.
    """
    callback(True)



"""Transformer API for laskea.

Typical filter data comes in as JSON and maps column keys to filter tasks:

{
  "order": ["keep", "drop", "replace"],
  "keep": [
    ["startswith", "ABC-"],
    ["contains", "Z"],
    ["icontains", "m"],
    ["equals", "DEF-42"],
    ["endswith", "-123"]
  ],
  "drop": [
    ["matches", "[A-Z]+-\\d+"]
  ],
  "replace": [
    ["DEF-", "definition-"]
  ]
}

"""
import re
from laskea import FILTER_MAP_TYPE, FILTER_ORDER_TYPE, FILTER_PAYLOAD_TYPE, log

# operation keys
DROP = 'drop'
KEEP = 'keep'
ORDER = 'order'
REPLACE = 'replace'

# action keys
CONTAINS = 'contains'
ENDSWITH = 'endswith'
EQUALS = 'equals'
ICONTAINS = 'icontains'
IENDSWITH = 'iendswith'
IEQUALS = 'iequals'
ISTARTSWITH = 'istartswith'
MATCHES = 'matches'
STARTSWITH = 'startswith'

ACTION_KEYS = (
    CONTAINS,
    ENDSWITH,
    EQUALS,
    ICONTAINS,
    IENDSWITH,
    IEQUALS,
    ISTARTSWITH,
    MATCHES,
    STARTSWITH,
)


def op_contains(entry: str, that: str) -> bool:
    """Case sensitive contains."""
    return bool(that in entry)


def op_endswith(entry: str, that: str) -> bool:
    """Case sensitive ends with."""
    return bool(entry.endswith(that))


def op_equals(entry: str, that: str) -> bool:
    """Case sensitive equals."""
    return bool(that == entry)


def op_icontains(entry: str, that: str) -> bool:
    """Case insensitive contains."""
    return bool(that.lower() in entry.lower())


def op_iendswith(entry: str, that: str) -> bool:
    """Case insensitive ends with."""
    return bool(entry.lower().endswith(that.lower()))


def op_iequals(entry: str, that: str) -> bool:
    """Case insensitive equals."""
    return bool(that.lower() == entry.lower())


def op_istartswith(entry: str, that: str) -> bool:
    """Case insensitive starts with."""
    return bool(entry.lower().startswith(that.lower()))


def op_matches(entry: str, that: str) -> bool:
    """Matches regular expression."""
    return bool(re.compile(that).match(entry))


def op_startswith(entry: str, that: str) -> bool:
    """Case sensitive starts with."""
    return bool(entry.startswith(that))


ACTION_MAP = {
    CONTAINS: op_contains,
    ENDSWITH: op_endswith,
    EQUALS: op_equals,
    ICONTAINS: op_icontains,
    IENDSWITH: op_iendswith,
    IEQUALS: op_iequals,
    ISTARTSWITH: op_istartswith,
    MATCHES: op_matches,
    STARTSWITH: op_startswith,
}


class FilterMap:
    """The class FilterMap validates the task data against known operations and required arguments.

    Known operations are:

    - drop
    - keep
    - replace

    A meta operation is:

    - order

    This "operation" is optional but if present must fully specify the order of application of the "real" operations.

    The default order of application is

    1. keep
    2. drop
    3. replace

    Real operation JSON member values are a list of list of strings (the payloads).
    The payloads have length two with semantics depending on the operation.

    Operations keep amd drop both iterate over all payloads in the order given
    by applying the action encoded in the first list item and using the second item as parameter
    on the cell content (list of strings) elements.

    The encoding of actions is as follows (for cell entry `entry` and payload parameter `that`):

    - contains - `that in entry`
    - endswith - `entry.endswith(that)`
    - equals - `that == entry`
    - icontains - `that.lower() in entry.lower()`
    - iendswith - `entry.lower().endswith(that.lower())`
    - iequals - `that.lower() == entry.lower()`
    - istartswith - `entry.lower().startswith(that.lower())`
    - matches - `re.compile(that).match(entry)`
    - startswith - `entry.startswith(that)`

    The third operation (replace) is delegated to the string replace function as action.
    In this case a payload pair like ["this", "with that"] is applied as `entry.replace('this', 'with that')`.

    """

    ORDER: FILTER_ORDER_TYPE = [KEEP, DROP, REPLACE]

    def __init__(self, column: str, filter_data: FILTER_MAP_TYPE):
        self.column = column
        self.filter_data: FILTER_MAP_TYPE = filter_data

        self.order = self.filter_data[ORDER] if self.filter_data.get(ORDER, []) else FilterMap.ORDER
        log.debug(f'{self.order=}')

        self.keeps: FILTER_PAYLOAD_TYPE = self.filter_data.get(KEEP, [])  # type: ignore
        self.drops: FILTER_PAYLOAD_TYPE = self.filter_data.get(DROP, [])  # type: ignore
        self.replaces: FILTER_PAYLOAD_TYPE = self.filter_data.get(REPLACE, [])  # type: ignore

        self.operations = []
        for kind in self.order:
            if kind == KEEP:
                if self.keeps:
                    self.operations.append((kind, self.keeps))
            elif kind == DROP:
                if self.drops:
                    self.operations.append((kind, self.drops))
            elif kind == REPLACE:
                if self.replaces:
                    self.operations.append((kind, self.replaces))
            else:
                log.warning(f'ignored order element ({kind}) - please verify your filter data')

    def apply(self, entry: str) -> str:
        """Initial naive application during stage 1 implementation of transformer."""
        if not entry.strip():  # TODO(sthagen) - this may exclude use cases of manipulating space ;-)
            return ''
        if not self.operations:
            return entry
        transformed = entry
        pre_replace = False
        if self.operations:
            kind, tasks = self.operations[0]
            if kind == REPLACE and tasks:
                pre_replace = True
                for this, with_that in tasks:
                    transformed = transformed.replace(this, with_that)

        for kind, tasks in self.operations:
            log.debug(f'+ applying kind ({kind})')
            if kind in (KEEP, DROP):
                if tasks:
                    for key, parameter in tasks:
                        log.debug(f'  - applying key ({key}) with parameter ({parameter}) for kind ({kind})')
                        if key.lower() not in ACTION_KEYS:
                            log.warning(f'skipping action with unknown key ({key}) for operation type ({kind})')
                            continue
                        hit = ACTION_MAP[key.lower()](transformed, parameter)
                        log.debug(f'    ==> {"hit" if hit else "miss"}')
                        if hit:
                            if kind == DROP:
                                return ''
                            if kind == KEEP:
                                return transformed
            elif not pre_replace:  # REPLACE
                if tasks:
                    for this, with_that in tasks:
                        transformed = transformed.replace(this, with_that)

        return transformed



import laskea.transform as tr


def test_op_contains_hits():
    assert tr.op_contains('entry', 'nt')


def test_op_contains_misses():
    assert not tr.op_contains('entry', 'NT')


def test_op_endswith_hits():
    assert tr.op_endswith('entry', 'try')


def test_op_endswith_misses():
    assert not tr.op_endswith('entry', 'different')


def test_op_equals_hits():
    assert tr.op_equals('entry', 'entry')


def test_op_equals_misses():
    assert not tr.op_equals('entry', 'different')


def test_op_icontains_hits():
    assert tr.op_icontains('entry', 'NT')


def test_op_icontains_misses():
    assert not tr.op_icontains('entry', 'different')


def test_op_iendswith_hits():
    assert tr.op_iendswith('entry', 'TRY')


def test_op_iendswith_misses():
    assert not tr.op_iendswith('entry', 'different')


def test_op_iequals_hits():
    assert tr.op_iequals('entry', 'ENTRY')


def test_op_iequals_misses():
    assert not tr.op_iequals('entry', 'different')


def test_op_istartswith_hits():
    assert tr.op_istartswith('entry', 'EN')


def test_op_istartswith_misses():
    assert not tr.op_istartswith('entry', 'different')


def test_op_matches_hits():
    assert tr.op_matches('entry', 'e.*')


def test_op_matches_misses():
    assert not tr.op_matches('entry', 'd.*')


def test_op_startswith_hits():
    assert tr.op_startswith('entry', 'en')


def test_op_startswith_misses():
    assert not tr.op_startswith('entry', 'different')


def test_init_class():
    c_filter = tr.FilterMap('c', {})
    assert c_filter.order == tr.FilterMap.ORDER


def test_apply_empty():
    c_filter = tr.FilterMap('c', {})
    assert c_filter.apply('') == ''


def test_apply_spaces_only():
    c_filter = tr.FilterMap('c', {})
    assert c_filter.apply(' ') == ''


def test_apply_no_filters():
    c_filter = tr.FilterMap('c', {})
    assert c_filter.apply('foo') == 'foo'


def test_apply_drop_that():
    c_filter = tr.FilterMap('c', {'drop': [['equals', 'that']]})
    assert c_filter.apply('that') == ''


def test_apply_drop_that_miss():
    c_filter = tr.FilterMap('c', {'drop': [['equals', 'that']]})
    assert c_filter.apply('else') == 'else'


def test_apply_keep_drop_that():
    c_filter = tr.FilterMap('c', {'keep': [['iequals', 'that']], 'drop': [['equals', 'THAT']]})
    assert c_filter.apply('THAT') == 'THAT'


def test_apply_replace_keep_drop_that():
    c_filter = tr.FilterMap(
        'c',
        {
            'order': ['replace', 'keep', 'drop'],
            'replace': [['THAT', 'that']],
            'keep': [['equals', 'THAT']],
            'drop': [['equals', 'that']],
        },
    )
    assert c_filter.apply('THAT') == ''


def test_apply_multiple_replace_keep_drop_that():
    c_filter = tr.FilterMap(
        'c',
        {
            'order': ['replace', 'keep', 'drop'],
            'replace': [['THAT', 'that'], ['that', 'SOMETHING_COMPLETELY_DIFFERENT']],
            'keep': [['equals', 'THAT']],
            'drop': [['equals', 'that']],
        },
    )
    assert c_filter.apply('THAT') == 'SOMETHING_COMPLETELY_DIFFERENT'
